from typing import TypeVar, Generic, Optional

from django.db.models import Model, Q, QuerySet
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext_lazy as _

from dry_django.selectors.generic import BaseModelSelector

_DjangoModelTypeVar = TypeVar("_DjangoModelTypeVar", bound=Model)


#################################
#################################
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#  DEPRECATED WILL BE REMOVED   #
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#################################
#################################


class DeprecatedUUIDSelectorMixin(BaseModelSelector[_DjangoModelTypeVar], Generic[_DjangoModelTypeVar]):
    @classmethod
    def get_by_uuid(
        cls, *, uuid: str, select_related: Optional[tuple[str]] = None, prefetch_related: Optional[tuple[str]] = None
    ) -> _DjangoModelTypeVar:
        query = Q(uuid=uuid)
        try:
            return cls.wrap_queryset_with_select_and_prefetch_related(
                cls.model.objects, select_related=select_related, prefetch_related=prefetch_related
            ).get(query)
        except cls.model.DoesNotExist:
            raise ObjectDoesNotExist(_("%s with uuid = %s not found.") % (cls.model.__name__, uuid))

    @classmethod
    def list_get_by_uuids(
        cls,
        *,
        uuids: list[str],
        select_related: Optional[tuple[str]] = None,
        prefetch_related: Optional[tuple[str]] = None,
    ) -> QuerySet:
        return cls.wrap_queryset_with_select_and_prefetch_related(
            cls.model.objects, select_related=select_related, prefetch_related=prefetch_related
        ).filter(uuid__in=uuids)



from django.core.exceptions import ValidationError as DjangoValidationError, PermissionDenied
from django.http import Http404

from rest_framework.views import exception_handler
from rest_framework import exceptions
from rest_framework.serializers import as_serializer_error
from rest_framework.response import Response
from rest_framework.settings import api_settings

from dry_django.exceptions import ApplicationError


class ErrorsFormatter:
    """
    Данный форматтер ошибок приводит ошибки в корректной форме,
    для этого он использует стандартные код и сообщение ошибки из DRF
    и приводит их к корректному формату
    {
        "errors": [
            {
                "message": "Error message",
                "code": "Some code",
                "field": "field_name"
            },
            {
                "message": "Error message",
                "code": "Some code",
                "field": "nested.field_name"
            },
            ...
        ]
    }
    """

    FIELD = "field"
    MESSAGE = "message"
    CODE = "code"
    ERRORS = "errors"

    def __init__(self, exception):
        self.exception = exception

    def __call__(self):
        if hasattr(self.exception, "get_full_details"):
            formatted_errors = self._get_response_json_from_drf_errors(
                serializer_errors=self.exception.get_full_details()
            )
        else:
            formatted_errors = self._get_response_json_from_error_message(message=str(self.exception))

        return formatted_errors

    def _get_response_json_from_drf_errors(self, serializer_errors=None):
        if serializer_errors is None:
            serializer_errors = {}

        if type(serializer_errors) is list:
            serializer_errors = {api_settings.NON_FIELD_ERRORS_KEY: serializer_errors}

        list_of_errors = self._get_list_of_errors(errors_dict=serializer_errors)

        response_data = {self.ERRORS: list_of_errors}

        return response_data

    def _get_response_json_from_error_message(self, *, message="", field=None, code="error"):
        response_data = {self.ERRORS: [{self.MESSAGE: message, self.CODE: code}]}

        if field:
            response_data[self.ERRORS][self.FIELD] = field

        return response_data

    def _unpack(self, obj):
        if type(obj) is list and len(obj) == 1:
            return obj[0]

        return obj

    def _get_list_of_errors(self, field_path="", errors_dict=None):
        """
        Словарь с ошибками следующего формата:
        {
            'field1': {
                'message': 'соообщение..'
                'code' 'код ошибки...'
            },
            'field2: ...'
        }
        """
        if errors_dict is None:
            return []

        message_value = errors_dict.get(self.MESSAGE, None)

        # Note: If 'message' is name of a field we don't want to stop the recursion here!
        if message_value is not None and (type(message_value) in {str, exceptions.ErrorDetail}):
            if field_path:
                errors_dict[self.FIELD] = field_path
            return [errors_dict]

        errors_list = []
        for key, value in errors_dict.items():
            new_field_path = "{0}.{1}".format(field_path, key) if field_path else key
            key_is_non_field_errors = key == api_settings.NON_FIELD_ERRORS_KEY

            if type(value) is list:
                current_level_error_list = []
                new_value = value

                for error in new_value:
                    # if the type of field_error is list we need to unpack it
                    field_error = self._unpack(error)

                    if not key_is_non_field_errors:
                        field_error[self.FIELD] = new_field_path

                    current_level_error_list.append(field_error)
            else:
                path = field_path if key_is_non_field_errors else new_field_path

                current_level_error_list = self._get_list_of_errors(field_path=path, errors_dict=value)

            errors_list += current_level_error_list

        return errors_list


def exception_errors_format_handler(exc, context):
    response = exception_handler(exc, context)

    # If unexpected error occurs (server error, etc.)
    if response is None:
        return response

    formatter = ErrorsFormatter(exc)

    response.data = formatter()

    return response


def unified_error_exception_handler(exc, ctx):
    """
    {
        "message": "Error message",
        "extra": {}
    }
    """
    if isinstance(exc, DjangoValidationError):
        exc = exceptions.ValidationError(as_serializer_error(exc))

    if isinstance(exc, Http404):
        exc = exceptions.NotFound()

    if isinstance(exc, PermissionDenied):
        exc = exceptions.PermissionDenied()

    response = exception_handler(exc, ctx)

    # If unexpected error occurs (server error, etc.)
    if response is None:
        if isinstance(exc, ApplicationError):
            data = {"message": exc.message, "extra": exc.extra}
            return Response(data, status=400)

        return response

    if isinstance(exc.detail, (list, dict)):
        response.data = {"detail": response.data}

    if isinstance(exc, exceptions.ValidationError):
        response.data["message"] = "Validation error"
        response.data["extra"] = {"fields": response.data["detail"]}
    else:
        response.data["message"] = response.data["detail"]
        response.data["extra"] = {}

    del response.data["detail"]

    return response



from typing import TypeVar, Generic, Optional, Union, Mapping

from django.db.models import Model, Q, QuerySet
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext_lazy as _
from dry_core.selectors.generics import Selector


_DjangoModelTypeVar = TypeVar("_DjangoModelTypeVar", bound=Model)


class ModelSelector(Selector[_DjangoModelTypeVar], Generic[_DjangoModelTypeVar]):
    select_related: Optional[Union[list, tuple]] = None
    prefetch_related: Optional[Union[list, tuple]] = None

    def __init__(
        self, select_related: Optional[Union[list, tuple]] = None, prefetch_related: Optional[Union[list, tuple]] = None
    ):
        self.select_related = select_related or self.select_related
        self.prefetch_related = prefetch_related or self.prefetch_related
        self.model_manager = self.model.objects

    def get(self, **kwargs) -> _DjangoModelTypeVar:
        filtered_kwargs = self._filter_kwargs_from_service_fields(kwargs)
        query = Q(**filtered_kwargs)
        try:
            return self.wrap_queryset_with_select_and_prefetch_related(self.model_manager, **kwargs).get(query)
        except self.model.DoesNotExist:
            raise ObjectDoesNotExist(_("%s with params: '%s' not found.") % (self.model.__name__, filtered_kwargs))

    async def aget(self, **kwargs) -> _DjangoModelTypeVar:
        filtered_kwargs = self._filter_kwargs_from_service_fields(kwargs)
        query = Q(**filtered_kwargs)
        try:
            return await self.wrap_queryset_with_select_and_prefetch_related(self.model_manager, **kwargs).aget(query)
        except self.model.DoesNotExist:
            raise ObjectDoesNotExist(_("%s with params: '%s' not found.") % (self.model.__name__, filtered_kwargs))

    def list_all(self, **kwargs) -> QuerySet:
        return self.wrap_queryset_with_select_and_prefetch_related(self.model_manager, **kwargs).all()

    def list_filter(self, **kwargs) -> QuerySet:
        return self.wrap_queryset_with_select_and_prefetch_related(self.model_manager, **kwargs).filter(
            **self._filter_kwargs_from_service_fields(kwargs)
        )

    ################################
    #       Utils & helpers
    ################################

    def wrap_queryset_with_select_and_prefetch_related(self, queryset: QuerySet, **kwargs) -> QuerySet:
        select_related_fields = self._get_select_related_fields_taking_into_kwargs(**kwargs)
        prefetch_related_fields = self._get_prefetch_related_fields_taking_into_kwargs(**kwargs)
        result_queryset = queryset
        if select_related_fields:
            result_queryset = queryset.select_related(*select_related_fields)
        if prefetch_related_fields:
            result_queryset = queryset.prefetch_related(*prefetch_related_fields)
        return result_queryset

    def _get_select_related_fields_taking_into_kwargs(self, **kwargs):
        return kwargs.get("select_related", self.select_related)

    def _get_prefetch_related_fields_taking_into_kwargs(self, **kwargs):
        return kwargs.get("prefetch_related", self.prefetch_related)

    @staticmethod
    def _filter_kwargs_from_service_fields(kwargs: Mapping) -> Mapping:
        return {key: value for key, value in kwargs if key not in ("select_related", "prefetch_related")}


#################################
#################################
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#  DEPRECATED WILL BE REMOVED   #
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#################################
#################################


class BaseModelSelector(Selector[_DjangoModelTypeVar], Generic[_DjangoModelTypeVar]):
    default_select_related: tuple = ()
    default_prefetch_fields: tuple = ()

    @classmethod
    def _get_select_related_fields(cls, select_related: Optional[tuple[str]] = None) -> tuple[str]:
        return select_related if select_related else cls.default_select_related

    @classmethod
    def _get_prefetch_related_fields(cls, prefetch_related: Optional[tuple[str]] = None) -> tuple[str]:
        return prefetch_related if prefetch_related else cls.default_prefetch_fields

    @classmethod
    def wrap_queryset_with_select_and_prefetch_related(
        cls,
        queryset: QuerySet,
        select_related: Optional[tuple[str]] = None,
        prefetch_related: Optional[tuple[str]] = None,
    ) -> QuerySet:
        return queryset.select_related(*cls._get_select_related_fields(select_related=select_related)).prefetch_related(
            *cls._get_prefetch_related_fields(prefetch_related=prefetch_related)
        )

    @classmethod
    def list_all(
        cls, *, select_related: Optional[tuple[str]] = None, prefetch_related: Optional[tuple[str]] = None
    ) -> QuerySet:
        return cls.wrap_queryset_with_select_and_prefetch_related(
            cls.model.objects, select_related=select_related, prefetch_related=prefetch_related
        ).all()

    @classmethod
    def get_by_id(
        cls, *, id: str, select_related: Optional[tuple[str]] = None, prefetch_related: Optional[tuple[str]] = None
    ) -> _DjangoModelTypeVar:
        query = Q(id=id)
        try:
            return cls.wrap_queryset_with_select_and_prefetch_related(
                cls.model.objects, select_related=select_related, prefetch_related=prefetch_related
            ).get(query)
        except cls.model.DoesNotExist:
            raise ObjectDoesNotExist(_("%s with id = %s not found.") % (cls.model.__name__, id))

    @classmethod
    def list_get_by_ids(
        cls,
        *,
        ids: list[int],
        select_related: Optional[tuple[str]] = None,
        prefetch_related: Optional[tuple[str]] = None,
    ) -> QuerySet:
        return cls.wrap_queryset_with_select_and_prefetch_related(
            cls.model.objects, select_related=select_related, prefetch_related=prefetch_related
        ).filter(id__in=ids)



from collections import OrderedDict
from typing import Type

from rest_framework import serializers
from rest_framework.pagination import LimitOffsetPagination as _LimitOffsetPagination
from rest_framework.response import Response

from logging import getLogger

logger = getLogger(__name__)


class LimitOffsetPagination(_LimitOffsetPagination):
    default_limit = 50
    max_limit = 200

    def get_paginated_data(self, data):
        return OrderedDict(
            [
                ("limit", self.limit),
                ("offset", self.offset),
                ("count", self.count),
                ("next", self.get_next_link()),
                ("previous", self.get_previous_link()),
                ("results", data),
            ]
        )

    def get_paginated_response(self, data):
        return Response(
            OrderedDict(
                [
                    ("limit", self.limit),
                    ("offset", self.offset),
                    ("count", self.count),
                    ("next", self.get_next_link()),
                    ("previous", self.get_previous_link()),
                    ("results", data),
                ]
            )
        )


def generate_paginated_response_serializer(
    instance_serializer: Type[serializers.Serializer], *, reference_name: str = None
):
    if reference_name is None:
        try:
            reference_name = instance_serializer.Meta.ref_name
        except AttributeError:
            reference_name = instance_serializer.__qualname__
    reference_name = "PaginatedResponse." + reference_name

    class ResponseSerializer(serializers.Serializer):
        limit = serializers.IntegerField(
            read_only=True,
        )
        offset = serializers.IntegerField(
            read_only=True,
        )
        count = serializers.IntegerField(
            read_only=True,
        )
        next = serializers.URLField(
            read_only=True,
        )
        previous = serializers.URLField(
            read_only=True,
        )
        results = serializers.ListSerializer(child=instance_serializer())

        class Meta:
            ref_name = reference_name

    return ResponseSerializer



from typing import Any, List, Iterable

from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


def only_one_field_set_validator(instance: Any, fields: List[str], raise_exception: bool = False) -> bool:
    """
    Проверяет, что instance имеет ровно одно установленное поле из fields,
    в случае если это не так генерирует ValidationError c нужным сообщением об ошибке
    :param instance:
    :param fields:
    :param raise_exception
    :return:
    """
    found_non_empty_field = False

    for field in fields:
        if getattr(instance, field, None) is not None:
            if found_non_empty_field:
                if raise_exception:
                    raise ValidationError(
                        _("Only one of %(fields)s must be set for %(instance_class)s")
                        % {"instance_class": type(instance).__name__, "fields": ", ".join(fields)}
                    )
                else:
                    return False
            else:
                found_non_empty_field = True

    if not found_non_empty_field:
        if raise_exception:
            raise ValidationError(
                _("At least one of %(fields)s must be set for %(instance_class)s")
                % {"instance_class": type(instance).__name__, "fields": ", ".join(fields)}
            )
        else:
            return False

    return True


def no_more_than_fields_set_validator(
    instance: Any, fields: List[str], count: int = 1, raise_exception: bool = False
) -> bool:
    """
    Проверяет, что instance имеет не больше count установленных полей из fields,
    в случае если это не так генерирует ValidationError c нужным сообщением об ошибке
    :param instance:
    :param fields:
    :param count:
    :param raise_exception
    :return:
    """
    found_non_empty_fields_count = 0

    for field in fields:
        if getattr(instance, field, None) is not None:
            found_non_empty_fields_count += 1
            if found_non_empty_fields_count > count:
                if raise_exception:
                    raise ValidationError(
                        _("No more than %(count)s of %(fields)s must be set for %(instance_class)s")
                        % {"instance_class": type(instance).__name__, "fields": ", ".join(fields), "count": count}
                    )
                else:
                    return False
    return True


def instances_field_value_validator(
    instances_list: Iterable[Any], field: str, value: Any, raise_exception: bool = True
) -> bool:
    """
    Проверяет, что каждый instance из списка содержит в себе определённое поле с определённым значением.
    В противном случае генерирует сообщение об ошибке ValidationError либо возвращает False
    :param instances_list:
    :param field:
    :param value:
    :param raise_exception:
    :return:
    """
    for instance in instances_list:
        instance_field = getattr(instance, field, None)
        if not instance_field or instance_field != value:
            if raise_exception:
                raise ValidationError(
                    _("Values of field %(field)s of instances in list are not the same as required")
                    % {
                        "field": field,
                    }
                )
            else:
                return False
    return True



from typing import Any, Optional, List, Dict, Set, Union

from rest_framework import serializers


def create_serializer_class(name, fields):
    return type(name, (serializers.Serializer,), fields)


def inline_serializer(*, fields, data=None, **kwargs):
    serializer_class = create_serializer_class(name="", fields=fields)

    if data is not None:
        return serializer_class(data=data, **kwargs)

    return serializer_class(**kwargs)


class QueryFieldsSetupMixin(serializers.Serializer):
    """
    Миксина для обеспечения управления полями сериалайзера для вывода на основе
    переменных query (get) запроса в адресной строке
    """

    def __init__(self, *args, fields: Optional[Union[tuple, list]] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._fields_to_serialize = fields

    def get_field_names(self, *args, **kwargs):
        fields_to_serialize = self._fields_to_serialize
        if fields_to_serialize is None:
            default_fields_config = getattr(self.Meta, "default_fields", None)
            if default_fields_config is not None:
                fields_to_serialize = tuple(default_fields_config)
        if fields_to_serialize is None:
            fields_to_serialize = super().get_field_names(*args, **kwargs)
        return self.update_fields_to_serialize_by_request_query_params(fields_to_serialize)

    def update_fields_to_serialize_by_request_query_params(self, fields_to_serialize: Union[list, tuple]) -> list[str]:
        if not self.check_serializer_context():
            return fields_to_serialize

        direct_fields_list = self.get_direct_fields_list()
        if direct_fields_list is not None:
            return self.get_validated_fields(direct_fields_list)
        else:
            expand_fields = self.get_fields_to_expand() or []
            exclude_fields = self.get_fields_to_exclude() or []

            new_fields = [field for field in fields_to_serialize if field not in exclude_fields]
            new_fields += [field for field in expand_fields if field not in new_fields]
            return self.get_validated_fields(new_fields)

    def get_serializer_available_fields(self) -> Set[str]:
        return set(list(getattr(self.Meta, "fields", [])))

    def get_validated_fields(self, fields: List[str]) -> List[str]:
        available_fields = self.get_serializer_available_fields()
        return [field for field in fields if field in available_fields]

    def get_request_query_params(self) -> Dict[str, Any]:
        return getattr(self.get_request_from_context(), "query_params", {})

    def get_field_from_query_param_value(self, query_param: str) -> Optional[List[str]]:
        query_param_value: str = self.get_request_query_params().get(query_param, None)
        if query_param_value is None:
            return None
        return [field.strip().lower() for field in query_param_value.split(",") if field.strip()]

    def get_fields_to_expand(self) -> Optional[List[str]]:
        return self.get_field_from_query_param_value(query_param="expand")

    def get_fields_to_exclude(self) -> Optional[List[str]]:
        return self.get_field_from_query_param_value(query_param="exclude")

    def get_direct_fields_list(self) -> Optional[List[str]]:
        return self.get_field_from_query_param_value(query_param="fields")

    def get_request_from_context(self):
        return self.context.get("request", None)

    def check_serializer_context(self) -> bool:
        if self.get_request_from_context():
            return True
        return False


class DefaultOutputSerializer(QueryFieldsSetupMixin, serializers.Serializer):
    pass


class DefaultOutputModelSerializer(QueryFieldsSetupMixin, serializers.ModelSerializer):
    """
    Обычный сериалийзер для вывода информации о сущности на основе модели
    """

    pass


class DefaultCreateSerializer(serializers.Serializer):
    pass


class DefaultSetSerializer(serializers.Serializer):
    pass


class DefaultUpdateSerializer(serializers.Serializer):
    pass



from typing import Generic, TypeVar

from django.db.models import Model
from dry_core.services import Service as _BaseService

from . import mixins


__all__ = [
    "ModelService",
]


_DjangoModelTypeVar = TypeVar("_DjangoModelTypeVar", bound=Model)


class ModelService(
    mixins.CreateServiceMixin[_DjangoModelTypeVar],
    mixins.UpdateServiceMixin[_DjangoModelTypeVar],
    mixins.DestroyServiceMixin[_DjangoModelTypeVar],
    mixins.ServiceModelInstanceMixin[_DjangoModelTypeVar],
    _BaseService[_DjangoModelTypeVar],
    Generic[_DjangoModelTypeVar],
):
    """
    Services for Django models with create, update, destroy operations
    """



def get_first_matching_attr(obj, *attrs, default=None):
    for attr in attrs:
        if hasattr(obj, attr):
            return getattr(obj, attr)

    return default


def get_error_message(exc):
    if hasattr(exc, "message_dict"):
        return exc.message_dict
    error_msg = get_first_matching_attr(exc, "message", "messages")

    if isinstance(error_msg, list):
        error_msg = ", ".join(error_msg)

    if error_msg is None:
        error_msg = str(exc)

    return error_msg



from rest_framework.views import APIView

from dry_django.contrib.rest_framework.mixins import (
    ApiDefaultMixin,
    ModelListAPIMixin,
    ModelRetrieveAPIMixin,
    ModelCreateAPIMixin,
    ModelUpdateAPIMixin,
    ModelDestroyAPIMixin,
)


class DefaultAPIView(ApiDefaultMixin, APIView):
    """
    Базовый класс для создания представлений, включающий миксины по умолчанию
    """

    pass


###################################
#  Generic Dry Django views
###################################


class ModelListAPIView(ModelListAPIMixin, DefaultAPIView):
    pass


class ModelListCreateAPIView(ModelListAPIMixin, ModelCreateAPIMixin, DefaultAPIView):
    pass


class ModelRetrieveUpdateDestroyAPIView(
    ModelRetrieveAPIMixin, ModelUpdateAPIMixin, ModelDestroyAPIMixin, DefaultAPIView
):
    pass



from functools import cached_property
from typing import Generic, TypeVar, Any, Iterable, Union

from django.db.models import Model
from dry_core import operation
from dry_core.services import ServiceInstanceMixin


__all__ = ["ServiceModelInstanceMixin", "CreateServiceMixin", "UpdateServiceMixin", "DestroyServiceMixin"]


_DjangoModelTypeVar = TypeVar("_DjangoModelTypeVar", bound=Model)


class ServiceModelInstanceMixin(ServiceInstanceMixin[_DjangoModelTypeVar], Generic[_DjangoModelTypeVar]):
    def __init__(self, *args, **kwargs):
        super(ServiceModelInstanceMixin, self).__init__(*args, **kwargs)

    @cached_property
    def _model_field_names(self) -> set[str]:
        return {field.name for field in self.model._meta.get_fields()}

    def _filter_kwargs_by_model_fields(self, kwargs: dict[str, Any]):
        return {key: value for key, value in kwargs.items() if key in self._model_field_names}

    def save(self, commit: bool = True) -> None:
        if self.instance is None:
            return
        if not commit:
            return
        self.instance.save()


class CreateServiceMixin(ServiceModelInstanceMixin[_DjangoModelTypeVar], Generic[_DjangoModelTypeVar]):
    @operation
    def create(self, *, commit: bool = True, **kwargs) -> _DjangoModelTypeVar:
        self.instance: _DjangoModelTypeVar = self.model(**self._filter_kwargs_by_model_fields(kwargs))
        self.save(commit=commit)
        return self.instance

    @operation
    async def acreate(self, *, commit: bool = True, **kwargs) -> _DjangoModelTypeVar:
        if commit:
            self.instance: _DjangoModelTypeVar = await self.model.objects.acreate(
                **self._filter_kwargs_by_model_fields(kwargs)
            )
        else:
            self.instance: _DjangoModelTypeVar = self.model(**self._filter_kwargs_by_model_fields(kwargs))
        return self.instance

    @operation
    def bulk_create(
        self, instances_data: Iterable[Union[_DjangoModelTypeVar, dict[str, Any]]]
    ) -> list[_DjangoModelTypeVar]:
        not_dict_instances = [instance for instance in instances_data if not isinstance(instance, dict)]
        prepared_from_dict_instances = [
            self.create(commit=False, **instance_data)
            for instance_data in instances_data
            if isinstance(instance_data, dict) and instances_data
        ]
        return self.model.objects.bulk_create(not_dict_instances + prepared_from_dict_instances)

    @operation
    async def abulk_create(
        self, instances_data: Iterable[Union[_DjangoModelTypeVar, dict[str, Any]]]
    ) -> list[_DjangoModelTypeVar]:
        not_dict_instances = [instance for instance in instances_data if not isinstance(instance, dict)]
        prepared_from_dict_instances = [
            await self.acreate(commit=False, **instance_data)
            for instance_data in instances_data
            if isinstance(instance_data, dict) and instances_data
        ]
        return await self.model.objects.abulk_create(not_dict_instances + prepared_from_dict_instances)


class UpdateServiceMixin(ServiceModelInstanceMixin[_DjangoModelTypeVar], Generic[_DjangoModelTypeVar]):
    @operation
    def update(self, *, commit: bool = True, **kwargs) -> _DjangoModelTypeVar:
        self.validate_instance_filled()
        for attr, value in self._filter_kwargs_by_model_fields(kwargs).items():
            setattr(self.instance, attr, value)
        self.save(commit=commit)
        return self.instance

    @operation
    async def aupdate(self, *, commit: bool = True, **kwargs) -> _DjangoModelTypeVar:
        self.validate_instance_filled()
        for attr, value in self._filter_kwargs_by_model_fields(kwargs).items():
            setattr(self.instance, attr, value)
        self.save(commit=commit)
        return self.instance


class DestroyServiceMixin(ServiceModelInstanceMixin[_DjangoModelTypeVar], Generic[_DjangoModelTypeVar]):
    @operation
    def destroy(self) -> _DjangoModelTypeVar:
        self.validate_instance_filled()
        destroyed_instance: _DjangoModelTypeVar = self.instance
        self.instance.delete()
        self.instance = None
        return destroyed_instance

    @operation
    async def adestroy(self) -> _DjangoModelTypeVar:
        self.validate_instance_filled()
        destroyed_instance: _DjangoModelTypeVar = self.instance
        self.instance.delete()
        self.instance = None
        return destroyed_instance



from typing import Any, Union, Type, Optional

from django.core.exceptions import (
    ObjectDoesNotExist,
    ValidationError,
)
from django.db import IntegrityError
from django.db.models import QuerySet, Model
from django.http import QueryDict
from django.utils.translation import gettext_lazy as _
from django_filters import FilterSet
from rest_framework import exceptions as rest_exceptions
from rest_framework.pagination import BasePagination
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import Serializer
from rest_framework import status as http_statuses

from dry_django.contrib.rest_framework.utils import get_error_message
from dry_django.contrib.rest_framework.pagination import LimitOffsetPagination
from dry_django.selectors.generic import ModelSelector
from dry_django.services import ModelService

DEFAULT_FILTERS_SET_CLASS_NAME = "FilterSet"
DEFAULT_PAGINATION_CLASS = LimitOffsetPagination


class ApiErrorsMixin:
    """
    Миксина для трансформации исключений Python и Django в соответствующие исключения DRF,
    иначе они будут возвращать 500 код ответа, что не желательно.

    Миксина предназначена для использования в связке с APIView.
    """

    expected_exceptions = {
        ObjectDoesNotExist: rest_exceptions.NotFound,
        IntegrityError: rest_exceptions.ValidationError,
        ValueError: rest_exceptions.ValidationError,
        ValidationError: rest_exceptions.ValidationError,
        PermissionError: rest_exceptions.PermissionDenied,
    }

    def handle_exception(self, exc):
        if isinstance(exc, tuple(self.expected_exceptions.keys())):
            drf_exception_class = self.expected_exceptions[exc.__class__]
            drf_exception = drf_exception_class(get_error_message(exc))

            return super().handle_exception(drf_exception)

        return super().handle_exception(exc)


class ApiFiltersMixin:
    """
    Миксина для уменьшения шаблонного кода при работе с фильтрацией данных
    в get запросах.
    Определяет метод get_validated_filters для представлений на основе APIView,
    который позволяет получить словарь с провалидированными фильтрами на основе
    параметров запроса (get параметров).
    """

    def get_filtered_queryset(
        self, queryset: QuerySet[Any], *, filterset_class: Optional[Union[str, Type[FilterSet]]] = None
    ) -> QuerySet[Any]:
        """
        Возвращает QuerySet с фильтрами на основе параметров запроса (get параметров).
        Для валидации используется filterset_class.

        :param queryset: QuerySet для фильтрации
        :param filterset_class: Класс или имя класса FilterSet, если используется
            имя класса, то по нему ищется класс сериалайзера объявленный внутри класса представления.

        :return: Отфильтрованный на базе фильтров QuerySet
        """
        if filterset_class is None:
            filterset_class = DEFAULT_FILTERS_SET_CLASS_NAME
        if not isinstance(filterset_class, str) and not issubclass(filterset_class, FilterSet):
            raise TypeError(_("filterset_class must be str or class inherited from" " django_filters.FilterSet"))

        if isinstance(filterset_class, str):
            filterset_class = getattr(self, filterset_class, None)
            if filterset_class is None:
                return queryset

        if filterset_class is None or self.request is None:
            return queryset

        if not issubclass(filterset_class, FilterSet):
            raise TypeError(_("filterset_class must be inherited from" " django_filters.FilterSet class"))

        filterset = filterset_class(queryset=queryset, data=self.request.query_params, request=self.request)
        return filterset.qs


class ApiPaginatedResponseMixin:
    """
    Миксина для обеспечения функционала по пагинации List запросов,
    выполняемых через метод get в представлениях на базе APIView
    """

    def get_paginated_response(
        self,
        queryset,
        *,
        pagination_class: Type[BasePagination] = None,
        serializer_class: Type[Serializer] = None,
        serializer_context: dict[str, Any] = None,
        filterset_class: Optional[Union[str, Type[FilterSet]]] = None,
        as_raw_data: bool = False,
    ) -> Response:
        if pagination_class is None:
            pagination_class = getattr(self, "Pagination", None)
            if pagination_class is None:
                pagination_class = DEFAULT_PAGINATION_CLASS
        if not issubclass(pagination_class, BasePagination):
            raise AttributeError(
                _("pagination_class must be inherited from" " rest_framework.pagination.BasePagination class.")
            )

        if serializer_class is None:
            serializer_class = getattr(self, "OutputSerializer", None)
            if serializer_class is None:
                raise AttributeError(
                    _(
                        "serializer_class set as None by default,"
                        " but view class don't implements OutputSerializer class."
                    )
                )
        if not issubclass(serializer_class, Serializer):
            raise AttributeError(
                _("serializer_class must be inherited from" " rest_framework.serializers.Serializer class.")
            )

        paginator = pagination_class()
        if getattr(self, "get_filtered_queryset", None) is not None:
            queryset = self.get_filtered_queryset(queryset=queryset, filterset_class=filterset_class)

        page = paginator.paginate_queryset(queryset=queryset, request=self.request, view=self)

        if serializer_context is None:
            serializer_context = {}
        serializer_context.setdefault("request", self.request)

        if page is not None:
            serializer = serializer_class(page, many=True, read_only=True, context=serializer_context)
            if as_raw_data:
                return paginator.get_paginated_data(serializer.data)
            return paginator.get_paginated_response(serializer.data)

        serializer = serializer_class(queryset, many=True, read_only=True, context=serializer_context)
        if as_raw_data:
            return serializer.data
        return Response(data=serializer.data)


class ApiDefaultMixin(ApiPaginatedResponseMixin, ApiFiltersMixin, ApiErrorsMixin):
    """
    Миксина с набором базовых миксин необходимого для большинства APIView  представлений
    """

    def get_request_data_as_list(self, request):
        request_data = request.data
        if isinstance(request_data, QueryDict):
            request_data = [element for element in request_data.values()]
        return request_data


##############################################
#   Generic Operations Mixins
##############################################


class ModelLogicBaseAPIMixin:
    selector_class: Type[ModelSelector]
    service_class: Type[ModelService]


class _ModelRetrieveBase(ModelLogicBaseAPIMixin):
    retrieve_serializer: Type[Serializer]

    def serialize_obj(self, obj: Model) -> dict:
        return self.retrieve_serializer(obj, read_only=True).data

    def get_object(self, request: Request, *args, **kwargs):
        return self.selector_class().get(pk=args[0])


class ModelListAPIMixin(ApiPaginatedResponseMixin, ModelLogicBaseAPIMixin):
    list_item_serializer: Optional[Type[Serializer]] = None

    def get(self, request, *args, **kwargs):
        return self.get_paginated_response(
            queryset=self.get_list_items_queryset(request, *args, **kwargs), serializer_class=self.list_item_serializer
        )

    def get_list_items_queryset(self, request: Request, *args, **kwargs):
        return self.selector_class().list_all()


class ModelRetrieveAPIMixin(ApiPaginatedResponseMixin, _ModelRetrieveBase, ModelLogicBaseAPIMixin):
    def get(self, request, *args, **kwargs):
        obj = self.get_object(request, *args, **kwargs)
        return Response(self.serialize_obj(obj))


class ModelCreateAPIMixin(ApiPaginatedResponseMixin, _ModelRetrieveBase, ModelLogicBaseAPIMixin):
    create_request_body_serializer: Type[Serializer]

    def post(self, request, *args, **kwargs):
        request_data = self.parse_request_body_for_create(request, *args, **kwargs)
        obj = self.perform_create(request, *args, request_data=request_data, **kwargs)
        return Response(self.serialize_obj(obj))

    def parse_request_body_for_create(self, request, *args, **kwargs) -> dict:
        data = request.data
        if data is None:
            return {}
        serializer = self.create_request_body_serializer(data=data)
        serializer.is_valid(raise_exception=True)
        return serializer.data

    def perform_create(self, request: Request, *args, request_data: dict, **kwargs) -> Model:
        service = self.service_class()
        service.create(**request_data)
        return service.instance


class ModelUpdateAPIMixin(ApiPaginatedResponseMixin, _ModelRetrieveBase, ModelLogicBaseAPIMixin):
    update_request_body_serializer: Type[Serializer]

    def put(self, request, *args, **kwargs):
        obj = self.get_object(request, *args, **kwargs)
        request_data = self.parse_request_body_for_update(request, *args, **kwargs)
        obj = self.perform_update(obj, request, *args, request_data=request_data, **kwargs)
        return Response(self.serialize_obj(obj), status=http_statuses.HTTP_200_OK)

    def parse_request_body_for_update(self, request, *args, **kwargs) -> dict:
        data = request.data
        if data is None:
            return {}
        serializer = self.update_request_body_serializer(data=data)
        serializer.is_valid(raise_exception=True)
        return serializer.data

    def perform_update(self, obj: Model, request: Request, *args, request_data: dict, **kwargs) -> Model:
        service = self.service_class(instance=obj)
        service.update(**request_data)
        return service.instance


class ModelDestroyAPIMixin(ApiPaginatedResponseMixin, _ModelRetrieveBase, ModelLogicBaseAPIMixin):
    def delete(self, request, *args, **kwargs):
        obj = self.get_object(request, *args, **kwargs)
        self.perform_destroy(obj, request, *args, **kwargs)
        return Response(status=http_statuses.HTTP_204_NO_CONTENT)

    def perform_destroy(self, obj: Model, request: Request, *args, **kwargs) -> None:
        service = self.service_class(instance=obj)
        service.destroy()
